""" Utility functions for *pysg*."""
import inspect
import math
from functools import wraps

from pysg.error import PyrrTypeError
import numpy as np


def pyrr_type_checker(var: object, var_type: type) -> object:
    """ Makes sure that pyrr types are used as input. Tries to cast and raises PyrrTypeError if not possible.

    Args:
        var: Input variable.
        var_type: Type of input variable.

    Returns:
        Variable in correct type. Otherwise raise PyrrTypeError exception.

    """
    if type(var) is not var_type:
        try:
            return var_type(var)
        except Exception as e:
            raise PyrrTypeError(var, "Unsupported type. Expected %s type as input" % var_type)
    return var


def parameters_as_angles_deg_to_rad(*args_to_convert):
    """Converts specific angle arguments to angles in radians.

    Used as a decorator to reduce duplicate code.

    Arguments are specified by their argument name.
    For example
    ::

        @parameters_as_angles_deg_to_rad('a', 'b', 'optional')
        def myfunc(a, b, *args, **kwargs):
            pass

        myfunc(20, [15,15], optional=[80,80,80])
    """

    def decorator(fn):
        # wraps allows us to pass the docstring back
        # or the decorator will hide the function from our doc generator

        try:
            getfullargspec = inspect.getfullargspec
        except AttributeError:
            getfullargspec = inspect.getargspec

        @wraps(fn)
        def wrapper(*args, **kwargs):
            # get the arguments of the function we're decorating
            fn_args = getfullargspec(fn)

            # convert any values that are specified
            # if the argument isn't in our list, just pass it through

            # convert the *args list
            # we zip the args with the argument names we received from
            # the inspect function
            args = list(args)
            for i, (k, v) in enumerate(zip(fn_args.args, args)):
                if k in args_to_convert and v is not None:
                    if type(v) == float or type(v) == int:
                        args[i] = math.radians(v)
                    else:
                        args[i] = np.radians(v)

            # convert the **kwargs dict
            for k, v in kwargs.items():
                if k in args_to_convert and v is not None:
                    if type(v) == float or type(v) == int:
                        kwargs[k] = math.radians(v)
                    else:
                        kwargs[k] = np.radians(v)

            # pass the converted values to our function
            return fn(*args, **kwargs)

        return wrapper

    return decorator



# -*- coding: utf-8 -*-
"""The root of the scene graph

All children added to this node can be rendered via a renderer.

"""
from pysg.constants import color
from pysg.light import PointLight
from pysg.node_3d import Node3D


class RenderLists:

    def __init__(self):
        """ Data object containing lights and geometry list for rendering. """
        self.geometry = list()
        self.point_lights = list()


class Scene(Node3D):

    def __init__(self, background_color: tuple = color.rgb["black"],
                 ambient_light: tuple = (0., 0., 0.), auto_update: bool = True):
        """ The scene object. Must always bee the root node of the scene graph.

        Args:
            background_color: The clear color of the scene.
            ambient_light: Light value which will be applied to all objects in scene.
            auto_update: If true the object transform will be updated automatically.
            Otherwise you have to do it manually.
        """
        super().__init__()
        self.auto_update = auto_update
        self.background_color = background_color
        self.ambient_light = ambient_light
        self._render_lists = RenderLists()

    @property
    def render_list(self) -> RenderLists:
        """ The current render list containing all the 3D objects and lights for rendering.

        Returns:
            RenderLists: A RenderLists object containing all elements for rendering.
        """
        return self._render_lists

    def add(self, node_3d: 'Node3D') -> None:
        """ Overrides base class of Node3D to add an objects to render or light list.

        Args:
            node_3d (Node3D): The child node which shall be added to the scene graph.
        """

        for n in (node_3d.get_leaf_nodes()):
            object_3d_type = type(n)
            if issubclass(object_3d_type, PointLight):
                if len(self.render_list.point_lights) < 1:
                    self.render_list.point_lights.append(n)
                else:
                    print('Warning! Not more than one point light in a scene is possible right now.')
            else:
                self.render_list.geometry.append(n)

        super(Scene, self).add(node_3d)

    def remove(self, node_3d: 'Node3D') -> None:
        """ Overrides base class of Node3D to removes an object from render or light list.

        Args:
            node_3d (Node3D): The child node which shall be removed.
        """

        for n in (node_3d.get_leaf_nodes()):
            if issubclass(type(n), PointLight):
                self.render_list.point_lights.remove(n)
            else:
                self.render_list.geometry.remove(n)

        super().remove(node_3d)

    def clear(self) -> None:
        """ Clears render lists and scene graph. """

        self.render_list.point_lights = list()
        self.render_list.geometry = list()
        self.children = list()



from unittest import TestCase

from pysg import CubeObject3D, PointLight, Scene
from pysg.testing import CustomAssertions


class TestScene(TestCase, CustomAssertions):
    def setUp(self):
        self.scene = Scene()

    def test_add_1(self):
        cube = CubeObject3D(1, 1, 1)
        self.scene.add(cube)
        self.assertEqual(self.scene.render_list.geometry[0], cube)

    def test_add_2(self):
        light = PointLight(color=(1, 1, 1))
        self.scene.add(light)
        self.assertEqual(self.scene.render_list.point_lights[0], light)
        light2 = PointLight(color=(1, 1, 1))
        self.scene.add(light2)
        # Only one point light per scene is possible
        self.assertEqual(len(self.scene.render_list.point_lights), 1)

    def test_remove(self):
        cube = CubeObject3D(1, 1, 1)
        self.scene.add(cube)
        self.scene.remove(cube)
        self.assertEqual(len(self.scene.render_list.geometry), 0)

    def test_clear(self):
        cube = CubeObject3D(1, 1, 1)
        self.scene.add(cube)
        self.scene.clear()
        self.assertEqual(len(self.scene.render_list.geometry), 0)
        self.assertEqual(len(self.scene.children), 0)



from unittest import TestCase

import numpy as np
from pyrr import Vector3, Quaternion

from pysg import Node3D
from pysg.testing import CustomAssertions


class TestNode3D(TestCase, CustomAssertions):
    def setUp(self):
        # Root
        #   |  \
        #  C_1, Child_2
        #        /  \
        #     C_2_1 C_2_2

        self.root = Node3D("root")
        self.child_1 = Node3D("child_1")
        self.child_2 = Node3D("child_2")
        self.child_2_1 = Node3D("child_2_1")
        self.child_2_2 = Node3D("child_2_2")
        self.root.add(self.child_1)
        self.root.add(self.child_2)
        self.child_2.add(self.child_2_1)
        self.child_2.add(self.child_2_2)

    def test_local_position_updates_world(self):
        self.child_1.local_position = [1, 0, 0]
        self.assertEqual(self.child_1.local_position, self.child_1.world_position)

    def test_init_world_position(self):
        self.root.world_position = Vector3([1, 0, 0])
        child = Node3D()
        child.local_position = Vector3([1, 0, 0])
        self.root.add(child)
        # Child local position is now relative to parent
        self.assertEqual(0, child.local_position[0])

    def test_position_change_hierarchy(self):
        self.root.local_position += 1
        self.assertEqual(self.root.local_position, self.child_2_2.world_position)
        self.child_2.local_position += 1
        self.assertEqual(self.child_2.world_position, self.child_2_2.world_position)
        self.assertEqual(self.root.world_position, self.child_1.world_position)

    def test_rotate_x_1(self):
        self.root.rotate_x(180)
        self.assertQuaternionAreEqual(self.root.local_quaternion, Quaternion([1, 0, 0, 0]))

    def test_rotate_x_2(self):
        self.root.rotate_x(360)
        self.assertQuaternionAreEqual(self.root.local_quaternion, Quaternion([0, 0, 0, 1]))

    def test_rotate_x_3(self):
        self.root.rotate_x(720)
        self.assertQuaternionAreEqual(self.root.local_quaternion, Quaternion([0, 0, 0, 1]))

    def test_rotate_y_1(self):
        self.root.rotate_y(180)
        self.assertQuaternionAreEqual(self.root.local_quaternion, Quaternion([0, 1, 0, 0]))

    def test_rotate_y_2(self):
        self.root.rotate_y(360)
        self.assertQuaternionAreEqual(self.root.local_quaternion, Quaternion([0, 0, 0, 1]))

    def test_rotate_z_1(self):
        self.root.rotate_z(180)
        self.assertQuaternionAreEqual(self.root.local_quaternion, Quaternion([0, 0, 1, 0]))

    def test_rotate_z_2(self):
        self.root.rotate_z(360)
        self.assertQuaternionAreEqual(self.root.local_quaternion, Quaternion([0, 0, 0, 1]))

    def test_rotate_z_3(self):
        self.root.rotate_z(720)
        self.assertQuaternionAreEqual(self.root.local_quaternion, Quaternion([0, 0, 0, 1]))

    def test_rotate_z_4(self):
        self.root.rotate_z(180, local_space=False)
        self.assertQuaternionAreEqual(self.root.local_quaternion, Quaternion([0, 0, 1, 0]))
        self.assertQuaternionAreEqual(self.root.world_quaternion, Quaternion([0, 0, 1, 0]))

    def test_euler_1(self):
        self.root.local_euler_angles = Vector3([180, 0, 0])
        self.assertQuaternionAreEqual(self.root.local_quaternion, Quaternion([1, 0, 0, 0]))

    def test_euler_2(self):
        self.child_1.local_euler_angles = Vector3([80, 40, 10])
        np.testing.assert_almost_equal(np.array(self.child_1.local_euler_angles), np.array([80, 40, 10]))

    def test_euler_3(self):
        self.child_2.local_euler_angles = Vector3([180, 0, 0])
        self.assertQuaternionAreEqual(self.child_2.local_quaternion, Quaternion([1, 0, 0, 0]))

    def test_euler_4(self):
        angles = Vector3([45, 78, 54])
        self.child_2.local_euler_angles = angles
        np.testing.assert_almost_equal(np.array(self.child_2.local_euler_angles), np.array(angles))

    def test_quaternion_1(self):
        self.root.local_quaternion = [0.5, 0.5, 0.5, 0.5]
        self.assertQuaternionAreEqual(self.child_2.world_quaternion, self.root.local_quaternion)

    def test_quaternion_2(self):
        self.root.local_quaternion = [-0.5, 0.5, 0.5, 0.5]
        self.assertQuaternionAreEqual(self.child_2.world_quaternion, self.root.local_quaternion)

    def test_quaternion_3(self):
        self.root.world_quaternion = [-0.5, 0.5, 0.5, 0.5]
        self.assertQuaternionAreEqual(self.child_2.world_quaternion, self.root.local_quaternion)

    def test_quaternion_4(self):
        self.root.world_quaternion = [1, 0, 0, 0]
        self.child_2.local_quaternion = [1, 0, 0, 0]
        self.assertQuaternionAreEqual(self.child_2.world_quaternion, Quaternion([0, 0, 0, 1]))

    def test_scale_1(self):
        self.root.scale = Vector3([1., 2., 3.])
        np.testing.assert_almost_equal(np.array(self.root.scale), np.array([1., 2., 3.]))

    def test_scale_2(self):
        self.root.scale *= 2.
        np.testing.assert_almost_equal(np.array(self.root.scale), np.array([2., 2., 2.]))

    def test_scale_3(self):
        self.root.scale *= -2.
        np.testing.assert_almost_equal(np.array(self.root.scale), np.array([-2., -2., -2.]))

    def test_position_1(self):
        self.root.local_position += 2.
        np.testing.assert_almost_equal(np.array(self.root.local_position), np.array([2., 2., 2.]))
        np.testing.assert_almost_equal(np.array(self.root.world_position), np.array([2., 2., 2.]))

    def test_position_2(self):
        self.root.local_position += 2.
        np.testing.assert_almost_equal(np.array(self.child_1.world_position), np.array([2., 2., 2.]))
        np.testing.assert_almost_equal(np.array(self.child_2_1.world_position), np.array([2., 2., 2.]))

    def test_add_1(self):
        new_child = Node3D("new_child")
        self.root.add(new_child)
        self.assertEqual(new_child.parent, self.root)

    def test_add_2(self):
        new_child = Node3D("new_child")
        new_child.local_position = Vector3([1, 2, 3])
        np.testing.assert_almost_equal(np.array(new_child.world_position), np.array([1., 2., 3.]))
        self.root.local_position = Vector3([1, 1, 1])
        self.root.add(new_child)
        np.testing.assert_almost_equal(np.array(new_child.local_position), np.array([0., 1., 2.]))
        np.testing.assert_almost_equal(np.array(new_child.world_position), np.array([1., 2., 3.]))

    def test_remove(self):
        new_child = Node3D("new_child")
        self.root.add(new_child)
        self.assertTrue(new_child in self.root.children)
        self.root.remove(new_child)
        self.assertEqual(new_child.parent, None)
        self.assertFalse(new_child in self.root.children)



from unittest import TestCase

import numpy as np
from pyrr import Quaternion, Vector3

from pysg.pyrr_extensions import quaternion_are_equal, euler_angles_to_quaternion, quaternion_to_euler_angles
from pysg.testing import CustomAssertions


class TestPyrrExtensions(TestCase, CustomAssertions):
    def test_quaternion_are_equal_1(self):
        self.assertTrue(quaternion_are_equal(Quaternion([1, 0, 0, 0]), Quaternion([-1, 0, 0, 0])))

    def test_quaternion_are_equal_2(self):
        self.assertFalse(quaternion_are_equal(Quaternion([0, 1, 0, 0]), Quaternion([-1, 0, 0, 0])))

    def test_quaternion_are_equal_3(self):
        self.assertTrue(quaternion_are_equal(Quaternion([1, 0, 0, 0]), Quaternion([1, 0, 0, 0])))

    def test_euler_to_quaternion_1(self):
        v = np.radians(Vector3([0, 0, 0]))
        self.assertQuaternionAreEqual(euler_angles_to_quaternion(v), (Quaternion([0, 0, 0, 1])))
        v = np.radians([0, 0, 180])
        self.assertQuaternionAreEqual(euler_angles_to_quaternion(v), (Quaternion([0, 0, 1, 0])))
        v = np.radians([0, 180, 0])
        self.assertQuaternionAreEqual(euler_angles_to_quaternion(v), (Quaternion([0, 1, 0, 0])))
        v = np.radians([180, 0, 0])
        self.assertQuaternionAreEqual(euler_angles_to_quaternion(v), (Quaternion([1, 0, 0, 0])))

    def test_euler_to_quaternion_2(self):
        v = np.radians([0, 0, 360])
        self.assertQuaternionAreEqual(euler_angles_to_quaternion(v), (Quaternion([0, 0, 0, 1])))

    def test_euler_to_quaternion_3(self):
        v = np.radians(Vector3([90, 70, 45]))
        e = euler_angles_to_quaternion(v)
        self.assertQuaternionAreEqual(e, Quaternion([0.6903455, 0.5963678, -0.1530459, 0.3799282]), epsilon=1e-5)

    def test_quaternion_to_euler_1(self):
        q = Quaternion([0, 0, 0, 1])
        e = quaternion_to_euler_angles(q)
        np.testing.assert_almost_equal(np.array(e), np.array([0., 0., 0.]))

    def test_quaternion_to_euler_2(self):
        q = Quaternion([0, 0, 1, 0])
        e = quaternion_to_euler_angles(q)
        # Rotation [180, 180, 0] and [0,0,180] are the same thing.
        np.testing.assert_almost_equal(np.array(e), np.array([180., 180., 0.]))

    def test_quaternion_to_euler_3(self):
        q = Quaternion([0, 1, 0, 0])
        e = quaternion_to_euler_angles(q)
        np.testing.assert_almost_equal(np.array(e), np.array([0., 180., 0.]))

    def test_quaternion_to_euler_4(self):
        q = Quaternion([1, 0, 0, 0])
        e = quaternion_to_euler_angles(q)
        np.testing.assert_almost_equal(np.array(e), np.array([180., 0., 0.]))

    def test_quaternion_to_euler_5(self):
        q = Quaternion([0.3711136, 0.0742227, 0.519559, 0.7660444])
        e = quaternion_to_euler_angles(q)
        np.testing.assert_almost_equal(np.array(e), np.array([69.4060045, -31.1935729, 58.3315948]), decimal=5)



# -*- coding: utf-8 -*-
""" All rendering related functions and classes. The ModernGL python library is used for all rendering functions
in *pysg*."""
import os

import moderngl as mgl

from pysg.camera import Camera
from pysg.geometry import create_cube, create_plane, create_icosahedron, create_circle, create_triangle, \
    create_cylinder, create_tetrahedral, create_pyramid
from pysg.object_3d import PlaneObject3D, IcosahedronObject3D, CubeObject3D, CircleObject3D, TriangleObject3D, \
    CylinderObject3D, TetrahedralObject3D, PyramidObject3D
from pysg.scene import Scene


class Renderer:

    def __init__(self, scene: Scene, camera: Camera):
        """Base class. All renderer implementations need to inherit form this class.

        Args:
            scene (Scene): Scene which shall be rendered.
            camera (Camera): Camera which is used to view scene.

        """
        self.scene = scene
        self.camera = camera
        self.ctx = None
        """Context: 
        The ModernGL context object which can be used to set some OpenGL related settings.
        For example the context can be used to deactivate backface culling:
            ::

               camera.ctx.disable(moderngl.CULL_FACE)
               
        See https://moderngl.readthedocs.io/en/stable/reference/context.html for more information.
        """

    def _create_vertex_array(self, vertices, indices, normals):
        vbo = self.ctx.buffer(vertices.astype('f4').tobytes())
        ibo = self.ctx.buffer(indices.astype('i4').tobytes())
        nbo = self.ctx.buffer(normals.astype('f4').tobytes())
        vao_content = [
            (vbo, '3f', 'in_vert'),
            (nbo, '3f', 'in_norm')
        ]
        return self.ctx.vertex_array(self.prog, vao_content, index_buffer=ibo)

    def _setup(self):
        """ Call this method from children as soon as context object was create.
        """
        self.ctx.enable(mgl.CULL_FACE)
        self.ctx.front_face = 'ccw'
        self.ctx.enable(mgl.DEPTH_TEST)
        shader_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'shader')
        self.prog = self.ctx.program(
            vertex_shader=open(os.path.join(shader_path, 'simple.vert')).read(),
            fragment_shader=open(os.path.join(shader_path, 'simple.frag')).read())
        self.object_color = self.prog['ObjectColor']
        self.ambient_light = self.prog['AmbientLight']
        self.point_light_position = self.prog['PointLightPosition']
        self.point_light_color = self.prog['PointLightColor']
        self.model_matrix = self.prog['ModelMatrix']
        self.view_projection_matrix = self.prog['ViewProjectionMatrix']
        self.model_size = self.prog['ModelSize']

        self.cube_vao = self._create_vertex_array(*create_cube())
        self.plane_vao = self._create_vertex_array(*create_plane())
        self.icosahedron_vao = self._create_vertex_array(*create_icosahedron())
        self.circle_vao = self._create_vertex_array(*create_circle())
        self.triangle_vao = self._create_vertex_array(*create_triangle())
        self.cylinder_vao = self._create_vertex_array(*create_cylinder())
        self.tetrahedral_vao = self._create_vertex_array(*create_tetrahedral())
        self.pyramid_vao = self._create_vertex_array(*create_pyramid())

    def _render(self) -> None:
        """ Call this method from subclasses to render all objects in the scene.
        """
        self.ctx.clear(*self.scene.background_color)

        # Update projection matrices
        if self.scene.auto_update:
            self.scene.update_world_matrix()

        if self.camera._parent is None:
            self.camera.update_world_matrix()

        view_projection_mat44 = self.camera.projection_matrix * self.camera.world_matrix.inverse
        self.view_projection_matrix.write(view_projection_mat44.astype('f4').tobytes())
        self.ambient_light.value = self.scene.ambient_light

        # TODO implement several light sources and other types
        if len(self.scene.render_list.point_lights) > 0:
            self.point_light_color.value = self.scene.render_list.point_lights[0].color
            self.point_light_position.value = tuple(self.scene.render_list.point_lights[0].world_position)

        # Render 3D geometries
        for object_3d in self.scene.render_list.geometry:
            self.model_matrix.write(object_3d.world_matrix.astype('f4').tobytes())
            self.object_color.value = object_3d.color
            self.model_size.value = object_3d.size
            if issubclass(type(object_3d), PlaneObject3D):
                self.plane_vao.render(mgl.TRIANGLES)
            elif issubclass(type(object_3d), IcosahedronObject3D):
                self.icosahedron_vao.render(mgl.TRIANGLES)
            elif issubclass(type(object_3d), CubeObject3D):
                self.cube_vao.render(mgl.TRIANGLES)
            elif issubclass(type(object_3d), CircleObject3D):
                self.circle_vao.render(mgl.TRIANGLE_FAN)
            elif issubclass(type(object_3d), TriangleObject3D):
                self.triangle_vao.render(mgl.TRIANGLES)
            elif issubclass(type(object_3d), CylinderObject3D):
                self.cylinder_vao.render(mgl.TRIANGLES)
            elif issubclass(type(object_3d), TetrahedralObject3D):
                self.tetrahedral_vao.render(mgl.TRIANGLES)
            elif issubclass(type(object_3d), PyramidObject3D):
                self.pyramid_vao.render(mgl.TRIANGLES)
            else:
                raise NotImplementedError(object_3d, "Renderer for object3D not implemented yet")

    def render(self) -> None:
        """ Base render function which needs to be implemented by sub-classes."""
        raise NotImplementedError()


class GLRenderer(Renderer):

    def __init__(self, scene: Scene, camera: Camera):
        """Render the scene to a given viewport.

        Args:
            scene (Scene): Scene which shall be rendered.
            camera (Camera): Camera which is used to view scene.
        """
        super().__init__(scene, camera)
        self.ctx = mgl.create_context()
        super()._setup()

        self.viewport = None
        """ tuple: Viewport is a tuple of size four (x, y, width, height). """

    def render(self) -> None:
        """ Set viewport and call _render() function of base class."""
        self.ctx.viewport = self.viewport
        super()._render()


class HeadlessGLRenderer(Renderer):

    def __init__(self, scene: Scene, camera: Camera, *, width: int, height: int):
        """Render the scene to a framebuffer which can be read to CPU memory to be used as an image.

        Args:
            scene (Scene): Scene which shall be rendered.
            camera (Camera): Camera which is used to view scene.
            width (float): Width of output image in pixel
            height (float): Height of output image in pixel
        """

        super().__init__(scene, camera)
        self.ctx = mgl.create_standalone_context()
        super()._setup()

        self.fbo = self.ctx.framebuffer(
            self.ctx.renderbuffer((width, height)),
            self.ctx.depth_renderbuffer((width, height)),
        )

    def render(self) -> None:
        """ Render the current scene and camera into a buffer.
        The buffer can later be returned with the current_image method.
        """
        self.fbo.use()
        super()._render()

    def current_image(self) -> bytes:
        """ Render the current scene and camera into a buffer.
        The buffer can later be returned with the 'current_image' method.

        Returns:
            bytes: The rendered byte array. Copy from vRAM to RAM.
        """
        return self.fbo.read(components=3, alignment=1)



from pyrr import Quaternion

from pysg.pyrr_extensions import quaternion_are_equal


class CustomAssertions:
    def assertQuaternionAreEqual(self, q1: Quaternion, q2: Quaternion, epsilon: float = 1e-12) -> None:
        if not quaternion_are_equal(q1, q2, epsilon=epsilon):
            raise AssertionError('Quaternion q1(%s) and q2(%s) are not equal!' % (q1, q2))



"""Define all exceptions that occur in pysg.
"""


class Error(Exception):
    """Base class for exceptions."""
    pass


class ParameterError(Error):
    """Exception raised for invalid camera parameter."""

    def __init__(self, expr, msg):
        self.expr = expr
        self.msg = msg


class PyrrTypeError(Error):
    """Exception raised for invalid pysg type."""

    def __init__(self, expr, msg):
        self.expr = expr
        self.msg = msg


""" Export all modules needed to use pysg."""
from __future__ import absolute_import

# Renderer
from .renderer import GLRenderer
from .renderer import HeadlessGLRenderer

# Camera
from .camera import PerspectiveCamera
from .camera import OrthographicCamera

# Light
from .light import PointLight

# Node3D
from .node_3d import Node3D

# 3D Objects
from .object_3d import CircleObject3D
from .object_3d import CubeObject3D
from .object_3d import CylinderObject3D
from .object_3d import IcosahedronObject3D
from .object_3d import PlaneObject3D
from .object_3d import PyramidObject3D
from .object_3d import TetrahedralObject3D
from .object_3d import TriangleObject3D

# Scene
from .scene import Scene

# Constants
from . import constants

from pysg.version import __version__



# -*- coding: utf-8 -*-
""" Lights can be added to the scene to illuminate the objects.
"""

from pysg.node_3d import Node3D


class Light(Node3D):

    def __init__(self, color: tuple, name: str = "Light") -> 'Light':
        """ Base class of all lights which can be added to the scene.
        """
        super().__init__(name)
        self.color = color


class PointLight(Light):

    def __init__(self, color: tuple, name: str = "PointLight") -> 'PointLight':
        """ Point light emits light in all directions from a single point.

        Args:
            color: The light intensity of light source.
        """
        super().__init__(color, name)



from unittest import TestCase

import numpy as np
from pyrr import Vector3

from pysg.error import PyrrTypeError
from pysg.testing import CustomAssertions
from pysg.util import pyrr_type_checker, parameters_as_angles_deg_to_rad


class TestUtil(TestCase, CustomAssertions):
    def test_pyrr_type_checker_fail(self):
        in_var = [2, 1, 2, 4]
        with self.assertRaises(PyrrTypeError):
            pyrr_type_checker(in_var, Vector3)

    def test_pyrr_type_checker_success(self):
        in_var = [2, 1, 2]
        checked_var = pyrr_type_checker(in_var, Vector3)
        self.assertEqual(Vector3, type(checked_var))

    def test_parameters_as_angles_deg_to_rad(self):
        @parameters_as_angles_deg_to_rad('angle')
        def test(angle, result):
            self.assertAlmostEqual(angle, result, delta=5)

        test(45, 0.785398)
        test(0, 0)
        test(-280, -4.88692)

    def test_parameters_as_angles_deg_to_rad_array(self):
        @parameters_as_angles_deg_to_rad('angle')
        def test(angle, result):
            np.testing.assert_almost_equal(np.array(angle), np.array(result), decimal=5)

        test(Vector3([45, 0, -280]), Vector3([0.785398, 0, -4.88692]))



# -*- coding: utf-8 -*-
""" This module contains all geometric 3D objects which can be added to a scene. All 3D objects inherit from
the Object3D base class. 3D objects are defined via a color, the object size, and a name."""

import pysg.constants.color
from pysg.node_3d import Node3D


class Object3D(Node3D):

    def __init__(self, name: str = "Object3D", color=pysg.constants.color.rgb['white']):
        """ A Object3D instances can be added to a scene and rendered.

        Args:
            name (str): Name of Object3D node.
        """

        super().__init__(name=name)
        self.size = (1, 1, 1)
        self.color = color


class CubeObject3D(Object3D):

    def __init__(self, width: float, height: float, depth: float, color=pysg.constants.color.rgb['white'],
                 name: str = "CubeObject"):
        """Creates a simple cube geometry

        Args:
            width (float): Width of cube.
            height (float): Height of cube.
            depth (float): Depth of cube.
            color (tuple): Color of 3D object.
            name (str): Name of object.
        """
        super().__init__(color=color, name=name)
        self.size = (width, height, depth)


class PlaneObject3D(Object3D):

    def __init__(self, width: float, height: float, color=pysg.constants.color.rgb['white'],
                 name: str = "PlaneObject"):
        """Creates a simple plane geometry.

        .. note:: Per default, only the front face of a plane gets rendered. If you want to also show the
                backface, you have to configure the moderngl render context. See :ref:`renderer` section
                for more information.

        Args:
            width (float): Width of plane.
            height (float): Height of plane.
            color (tuple): Color of 3D object.
            name (str): Name of object.
        """
        super().__init__(color=color, name=name)
        self.size = (width, 1, height)


class CircleObject3D(Object3D):

    def __init__(self, radius: float, color=pysg.constants.color.rgb['white'],
                 name: str = "CircleObject"):
        """Creates a simple circle geometry.

        .. note:: Per default, only the front face of a plane gets rendered. If you want to also show the
                backface, you have to configure the moderngl render context. See :ref:`renderer` section
                for more information.

        Args:
            radius (float): Radius of circle.
            color (tuple): Color of 3D object.
            name (str): Name of object.
        """
        super().__init__(color=color, name=name)
        self.size = (radius, 1, radius)


class IcosahedronObject3D(Object3D):

    def __init__(self, radius: float, color=pysg.constants.color.rgb['white'],
                 name: str = "IcosahedronObject"):
        """Creates a icosahedron geometry.

        Args:
            radius (float): Radius of icosahedron.
            color (tuple): Color of 3D object.
            name (str): Name of object.
        """
        super().__init__(color=color, name=name)
        self.size = (radius, radius, radius)


class TriangleObject3D(Object3D):

    def __init__(self, width: float, height: float, color=pysg.constants.color.rgb['white'],
                 name: str = "TriangleObject"):
        """Creates a icosahedron geometry.

        Args:
            width (float): Width of triangle.
            height (float): Height of triangle.
            color (tuple): Color of 3D object.
            name (str): Name of object.
        """
        super().__init__(color=color, name=name)
        self.size = (width, 1, height)


class CylinderObject3D(Object3D):

    def __init__(self, height: float, radius: float, color=pysg.constants.color.rgb['white'],
                 name: str = "CylinderObject"):
        """Creates a icosahedron geometry.

        Args:
            width (float): Width of triangle.
            height (float): Height of triangle.
            color (tuple): Color of 3D object.
            name (str): Name of object.
        """
        super().__init__(color=color, name=name)
        self.size = (radius * 2, height, radius * 2)


class TetrahedralObject3D(Object3D):

    def __init__(self, radius: float, color=pysg.constants.color.rgb['white'],
                 name: str = "BoxObject"):
        """Creates a tetrahedral geometry.

        Args:
            radius (float): radius of edge points lying on unit sphere.
            color (tuple): Color of 3D object.
            name (str): Name of object.
        """
        super().__init__(color=color, name=name)
        self.size = (radius, radius, radius)


class PyramidObject3D(Object3D):

    def __init__(self, base_size: float, height: float, color=pysg.constants.color.rgb['white'],
                 name: str = "PyramidObject"):
        """Creates a square base pyramide geometry.

        Args:
            base_size (float): Size of the pyramid square base.
            height (float): Pyramid height.
            color (tuple): Color of 3D object.
            name (str): Name of object.
        """
        super().__init__(color=color, name=name)
        self.size = (base_size, height, base_size)



# -*- coding: utf-8 -*-
""" Create basic geometries which are used to create buffered primitives in vRAM."""
import math
from typing import Tuple

import numpy as np


def create_cube(dtype='float32') -> Tuple[np.array, np.array, np.array]:
    """ Create standard cube of size one.

    Args:
        dtype: Data type of output numpy array.

    Returns:
        Tuple[np.array,np.array,np.array]: Tuple of size 3. First is np array for vertices, second for indices,
        and last for the normals.

    """
    # half dimension
    width = 0.5
    height = 0.5
    depth = 0.5

    vertices = np.array([
        # front
        # top right
        (width, height, depth),
        # top left
        (-width, height, depth),
        # bottom left
        (-width, -height, depth),
        # bottom right
        (width, -height, depth),

        # right
        # top right
        (width, height, -depth),
        # top left
        (width, height, depth),
        # bottom left
        (width, -height, depth),
        # bottom right
        (width, -height, -depth),

        # back
        # top right
        (-width, height, -depth),
        # top left
        (width, height, -depth),
        # bottom left
        (width, -height, -depth),
        # bottom right
        (-width, -height, -depth),

        # left
        # top right
        (-width, height, depth),
        # top left
        (-width, height, -depth),
        # bottom left
        (-width, -height, -depth),
        # bottom right
        (-width, -height, depth),

        # top
        # top right
        (width, height, -depth),
        # top left
        (-width, height, -depth),
        # bottom left
        (-width, height, depth),
        # bottom right
        (width, height, depth),

        # bottom
        # top right
        (width, -height, depth),
        # top left
        (-width, -height, depth),
        # bottom left
        (-width, -height, -depth),
        # bottom right
        (width, -height, -depth),
    ], dtype=dtype)

    # For triangle type counter clockwise
    # top right -> top left -> bottom left
    # top right -> bottom left -> bottom right
    indices = np.tile(np.array([0, 1, 2, 0, 2, 3], dtype='int'), (6, 1))
    for face in range(6):
        indices[face] += (face * 4)
    indices.shape = (-1,)

    normals = np.array([
        # front
        (0, 0, 1,),
        (0, 0, 1,),
        (0, 0, 1,),
        (0, 0, 1,),

        # right
        (1, 0, 0,),
        (1, 0, 0,),
        (1, 0, 0,),
        (1, 0, 0,),

        # back
        (0, 0, -1,),
        (0, 0, -1,),
        (0, 0, -1,),
        (0, 0, -1,),

        # left
        (-1, 0, 0,),
        (-1, 0, 0,),
        (-1, 0, 0,),
        (-1, 0, 0,),

        # top
        (0, 1, 0,),
        (0, 1, 0,),
        (0, 1, 0,),
        (0, 1, 0,),

        # bottom
        (0, -1, 0,),
        (0, -1, 0,),
        (0, -1, 0,),
        (0, -1, 0,),
    ], dtype=dtype)

    return vertices, indices, normals


def create_icosahedron(dtype='float32') -> Tuple[np.array, np.array, np.array]:
    """ Create icosahedron geometry with radius one.
    seealso:: http://www.songho.ca/opengl/gl_sphere.html

    Args:
        dtype: Data type of output numpy array.

    Returns:
        Tuple[np.array,np.array,np.array]: Tuple of size 3. First is np array for vertices, second for indices,
        and last for the normals.

    """
    # Fixed radius of 1
    RADIUS = 1.

    h_angle_steps = math.pi / 180 * 72  # 72 degree = 360 / 5
    v_angle_steps = math.atan(1. / 2.)  # elevation = 26.565 degree

    vertices = np.zeros((60, 3), dtype=dtype)  # array of 60 vertices (20 triangles)
    h_angle_1st_row = -math.pi / 2. - h_angle_steps / 2.  # start from -126 deg at 1st row
    h_angle_2nd_row = -math.pi / 2.  # start from -90 deg at 2nd row
    normals = np.zeros((60, 3), dtype=dtype)

    # Top vertex at(0, 0, r)
    v_top = np.array([0, 0, RADIUS])

    # 10 vertices at 1st and 2nd rows
    z = RADIUS * math.sin(v_angle_steps)  # elevation
    xy = RADIUS * math.cos(v_angle_steps)  # length on XY plane
    v_1st_row = np.zeros((5, 3))
    v_2nd_row = np.zeros((5, 3))
    for idx in range(0, 5):
        x_1 = xy * math.cos(h_angle_1st_row)
        x_2 = xy * math.cos(h_angle_2nd_row)
        y_1 = xy * math.sin(h_angle_1st_row)
        y_2 = xy * math.sin(h_angle_2nd_row)
        v_1st_row[idx] = np.array([x_1, y_1, z])
        v_2nd_row[idx] = np.array([x_2, y_2, -z])

        # next horizontal angles
        h_angle_1st_row += h_angle_steps
        h_angle_2nd_row += h_angle_steps

    # Bottom vertex at (0, 0, -r)
    v_bottom = np.array([0., 0., -RADIUS])

    # Helper function
    def set_normals(v_idx):
        v1 = vertices[v_idx] - vertices[v_idx + 1]
        v2 = vertices[v_idx] - vertices[v_idx + 2]
        normals[v_idx: v_idx + 2] = np.cross(v1, v2)

    # Set vertices and normals
    for idx in range(0, 5):
        # Top
        v_idx = idx * 3
        next_idx = (idx + 1) % 5
        vertices[v_idx] = v_top
        vertices[v_idx + 1] = v_1st_row[idx]
        vertices[v_idx + 2] = v_1st_row[next_idx]
        set_normals(v_idx)

        # First row
        v_idx = idx * 3 + (5 * 3)
        vertices[v_idx] = v_1st_row[next_idx]
        vertices[v_idx + 1] = v_1st_row[idx]
        vertices[v_idx + 2] = v_2nd_row[idx]
        set_normals(v_idx)

        # Second row
        v_idx = idx * 3 + (10 * 3)
        vertices[v_idx] = v_2nd_row[idx]
        vertices[v_idx + 1] = v_2nd_row[next_idx]
        vertices[v_idx + 2] = v_1st_row[next_idx]
        set_normals(v_idx)

        # Bottom
        v_idx = idx * 3 + (15 * 3)
        vertices[v_idx] = v_bottom
        vertices[v_idx + 1] = v_2nd_row[next_idx]
        vertices[v_idx + 2] = v_2nd_row[idx]
        set_normals(v_idx)

    indices = np.arange(0, 60, dtype='int')

    return vertices, indices, normals


def create_plane(dtype='float32') -> Tuple[np.array, np.array, np.array]:
    """ Create standard plane of size one.

    Args:
        dtype: Data type of output numpy array.

    Returns:
        Tuple[np.array,np.array,np.array]: Tuple of size 3. First is np array for vertices, second for indices,
        and last for the normals.

    """
    # half dimension
    width = 0.5
    height = 0.5

    vertices = np.array([
        # top right
        (width, 0.0, -height),
        # top left
        (-width, 0.0, -height),
        # bottom left
        (-width, 0.0, height),
        # bottom right
        (width, 0.0, height),
    ], dtype=dtype)

    # For triangle type counter clockwise
    # top right -> top left -> bottom left
    # top right -> bottom left -> bottom right
    indices = np.array([0, 1, 2, 0, 2, 3], dtype='int')

    normals = np.array([
        (0, 1, 0,),
        (0, 1, 0,),
        (0, 1, 0,),
        (0, 1, 0,)
    ], dtype=dtype)

    return vertices, indices, normals


def create_circle(dtype='float32', radius=1., fan_vertices=40) -> Tuple[np.array, np.array, np.array]:
    """ Create standard circle with radius one.

    Args:
        radius: Radius of circle.
        fan_vertices: Number of vertices used for triangle fan.
        dtype: Data type of output numpy array.

    Returns:
        Tuple[np.array,np.array,np.array]: Tuple of size 3. First is np array for vertices, second for indices,
        and last for the normals.

    """

    vertices = np.zeros((1 + fan_vertices, 3), dtype=dtype)
    vertices[0] = (0., 0., 0.)

    angle_step = (2 * math.pi) / fan_vertices
    angle = 0
    for idx in range(1, fan_vertices + 1):
        x = math.cos(angle) * radius
        y = math.sin(angle) * radius
        vertices[idx] = (x, 0., y)
        angle += angle_step

    indices = np.arange(0, 1 + fan_vertices, dtype='int')[::-1]

    normals = np.array([(0, 1, 0,), ] * (fan_vertices + 1), dtype=dtype)

    return vertices, indices, normals


def create_triangle(dtype='float32') -> Tuple[np.array, np.array, np.array]:
    """ Create standard triangle with side length one.

    Args:
        dtype: Data type of output numpy array.

    Returns:
        Tuple[np.array,np.array,np.array]: Tuple of size 3. First is np array for vertices, second for indices,
        and last for the normals.

    """

    h = 0.5 * math.sqrt(3)
    inner_circle_radius = math.sqrt(3) / 6.
    vertices = np.array([
        (0, 0, h - inner_circle_radius),
        (0.5, 0, -inner_circle_radius),
        (-0.5, 0, -inner_circle_radius),
    ], dtype=dtype)

    indices = np.arange(0, 3, dtype='int')

    normals = np.array([(0, 1, 0,), ] * 3, dtype=dtype)

    return vertices, indices, normals


def create_cylinder(dtype='float32') -> Tuple[np.array, np.array, np.array]:
    """ Create standard cylinder with height two and radius one.

    Args:
        dtype: Data type of output numpy array.

    Returns:
        Tuple[np.array,np.array,np.array]: Tuple of size 3. First is np array for vertices, second for indices,
        and last for the normals.

    """

    height = 2.
    radius = 1.
    sides = 6

    # Top and bottom share one center vertices and the triangles form a fan.
    # Each sides needs two unique triangle to render correct normals
    # Vertices layout: (top (1), upper_circle (sides), middle (4*sides)  ,lower_circle (sides), bottom (1).
    vertices = np.zeros((sides * 6 + 2, 3), dtype=dtype)
    normals = np.zeros(vertices.shape, dtype=dtype)
    # Every side has 4 triangles (two for middle, one for top, and one for bottom).
    indices = np.zeros((sides * 4, 3), dtype='int')

    y = height / 2.
    vertices[0] = (0., y, 0.)
    normals[0] = (0, 1, 0)
    vertices[-1] = (0., -y, 0.)
    normals[-1] = (0, -1, 0)
    angle_step = (2 * math.pi) / sides
    angle = 0
    for idx in range(1, sides + 1):
        x = math.cos(angle) * radius
        z = math.sin(angle) * radius
        # Top circle
        vertices[idx] = (x, y, z)
        normals[idx] = (0, 1, 0)
        # Bottom circle
        vertices[idx + (sides * 5)] = (x, -y, z)
        normals[-idx - 1] = (0, -1, 0)
        angle += angle_step

    # Top indices
    indices[0:sides] = [(0, (i + 1) % sides + 1, i + 1) for i in range(sides)]
    # Bottom indices
    offset = len(vertices) - 1
    indices[-sides:] = [(offset, offset - sides + i, offset - sides + (i + 1) % sides) for i in range(sides)]

    for idx in range(0, sides):
        array_idx = sides + idx * 4 + 1
        top_left = vertices[idx + 1]
        next_idx_top = idx + 2 if idx + 1 < sides else 1
        top_right = vertices[next_idx_top]
        bottom_left = vertices[idx - sides - 1]
        next_idx_bottom = idx - sides if idx - sides <= -2 else -sides - 1
        bottom_right = vertices[next_idx_bottom]

        vertices[array_idx] = top_left
        vertices[array_idx + 1] = top_right
        vertices[array_idx + 2] = bottom_left
        vertices[array_idx + 3] = bottom_right
        v1 = top_right - top_left
        v2 = bottom_left - top_left
        normal = np.cross(v1, v2) / np.linalg.norm(np.cross(v1, v2))
        normals[array_idx: (array_idx + 4)] = normal

        indices[sides + idx] = (array_idx, array_idx + 1, array_idx + 2)
        indices[sides * 2 + idx] = (array_idx + 1, array_idx + 3, array_idx + 2)

    indices = indices.flatten()

    return vertices, indices, normals


def create_tetrahedral(dtype='float32') -> Tuple[np.array, np.array, np.array]:
    """ Create tetrahedral geometry with radius one.

    Args:
        dtype: Data type of output numpy array.

    Returns:
        Tuple[np.array,np.array,np.array]: Tuple of size 3. First is np array for vertices, second for indices,
        and last for the normals.

    """
    size = 0.5

    v1 = np.array((size, size, size))
    v2 = np.array((size, -size, -size))
    v3 = np.array((-size, size, -size))
    v4 = np.array((-size, -size, size))

    vertices = np.array([
        # 1
        v4,
        v3,
        v2,

        # 2
        v3,
        v4,
        v1,

        # 3
        v1,
        v4,
        v2,

        # 4
        v2,
        v3,
        v1,
    ], dtype=dtype)

    norm_1 = tuple(np.cross((v4 - v2), (v3 - v2)))
    norm_2 = tuple(np.cross((v3 - v1), (v4 - v1)))
    norm_3 = tuple(np.cross((v4 - v1), (v2 - v1)))
    norm_4 = tuple(np.cross((v2 - v1), (v3 - v1)))

    normals = np.array([
        norm_1 * 3,
        norm_2 * 3,
        norm_3 * 3,
        norm_4 * 3,
    ])

    indices = np.arange(0, 12, dtype='int')

    return vertices, indices, normals


def create_pyramid(dtype='float32') -> Tuple[np.array, np.array, np.array]:
    """ Create regular pyramid geometry with square base with base size and height one.

    Args:
        dtype: Data type of output numpy array.

    Returns:
        Tuple[np.array,np.array,np.array]: Tuple of size 3. First is np array for vertices, second for indices,
        and last for the normals.

    """
    base_height = -0.333333

    tip_vert = np.array((0, 0.666666, 0))
    base_top_right_vert = np.array((0.5, base_height, 0.5))
    base_top_left_vert = np.array((-0.5, base_height, 0.5))
    base_bottom_right_vert = np.array((0.5, base_height, -0.5))
    base_bottom_left_vert = np.array((-0.5, base_height, -0.5))

    vertices = np.array([
        # Bottom
        base_top_right_vert,
        base_top_left_vert,
        base_bottom_left_vert,
        base_bottom_right_vert,

        # Front
        tip_vert,
        base_bottom_right_vert,
        base_bottom_left_vert,

        # Back
        tip_vert,
        base_top_left_vert,
        base_top_right_vert,

        # Right
        tip_vert,
        base_top_right_vert,
        base_bottom_right_vert,

        # Left
        tip_vert,
        base_bottom_left_vert,
        base_top_left_vert,
    ], dtype=dtype)

    norm_back = tuple(np.cross((base_top_left_vert - tip_vert), (base_top_right_vert - tip_vert)))
    norm_front = tuple(np.cross((base_bottom_right_vert - tip_vert), (base_bottom_left_vert - tip_vert)))
    norm_right = tuple(np.cross((base_top_right_vert - tip_vert), (base_bottom_right_vert - tip_vert)))
    norm_left = tuple(np.cross((base_bottom_left_vert - tip_vert), (base_top_left_vert - tip_vert)))

    normals = np.concatenate([
        (0, -1, 0) * 4,  # Bottom
        norm_front * 3,  # Front
        norm_back * 3,  # Back
        norm_right * 3,  # Right
        norm_left * 3  # Left
    ]).flatten()

    bottom_indices = np.array([0, 1, 2, 0, 2, 3])
    indices = np.concatenate([bottom_indices, np.arange(4, 16, dtype='int')])

    return vertices, indices, normals



""" Contains some functions which extend the pyrr math lib."""
import math

import numpy as np
from pyrr import Matrix44, Quaternion, Vector3
from pyrr.utils import parameters_as_numpy_arrays


def is_angle(angle, *, in_degrees, allow_negative, limit_to_circle):
    """ Check whether the given number is a valid angle rotation

    Args:
        angle (float): Number that needs to be evaluated.
        in_degrees (bool): If True expect number in degrees. Otherwise expect radians.
        allow_negative (bool): If False, no negative angles are allowed.
        limit_to_circle (bool): Determine whether the angle should be within a full circle.

    Returns:
        bool: True for valid angle, False otherwise.

    """
    if limit_to_circle:
        upper_limit = 360.0 if in_degrees else 2 * math.pi
        lower_limit = -360.0 if allow_negative else 0.0
    else:
        upper_limit = math.inf
        lower_limit = -math.inf if allow_negative else 0.0
    return lower_limit <= angle <= upper_limit


def compose_matrix(position: Vector3, quaternion: Quaternion, scale: Vector3) -> Matrix44:
    """ Reconstruct a 4x4 matrix from rotation, position and scale.

    Note that all matrices are in the OpenGL column first layout.

    Args:
        quaternion: Input rotation in Quaternion representation.
        position: Position in 3D space as Vector3.
        scale: Scale of object (x, y, and z) as Vector3.
    """

    translation_matrix = Matrix44([[1., 0., 0., 0.],
                                   [0., 1., 0., 0.],
                                   [0., 0., 1., 0.],
                                   [position[0], position[1], position[2], 1.]])
    rotation_matrix = quaternion.matrix44
    scale_matrix = Matrix44([[scale[0], 0., 0., 0.0],
                             [0., scale[1], 0., 0.0],
                             [0., 0., scale[2], 0.0],
                             [0., 0., 0., 1.]])
    return translation_matrix * rotation_matrix * scale_matrix


def quaternion_are_equal(q1: Quaternion, q2: Quaternion, epsilon: float = 1e-12) -> bool:
    """ Check whether two quaternions represent the same rotation.

    Note that q = -q

    Args:
        epsilon (float): Wiggle factor for rounding inaccuracy.
        q1: First quaternion.
        q2: Second quaternion.
    Returns:
        bool: True if both are equal. False if not.
    """

    return math.fabs(q1.dot(q2)) > 1 - epsilon


# From http://www.euclideanspace.com/maths/geometry/rotations/conversions/quaternionToAngle/
def quaternion_to_euler_angles(quaternion: Quaternion) -> Vector3:
    """ Converts quaternion to euler angles in degrees and YZX order.

    First heading, then attitude, then bank. See also:
    http://www.euclideanspace.com/maths/geometry/rotations/euler/index.htm
    and:
    http://www.euclideanspace.com/maths/geometry/rotations/conversions/quaternionToEuler/index.htm

    Returns:
        Vector3: Vector3 representation of quaternion
    """
    sqw = quaternion.w * quaternion.w
    sqx = quaternion.x * quaternion.x
    sqy = quaternion.y * quaternion.y
    sqz = quaternion.z * quaternion.z
    unit = sqx + sqy + sqz + sqw  # if normalised is one, otherwise is correction factor
    test = quaternion.x * quaternion.y + quaternion.z * quaternion.w
    if test > 0.499 * unit:  # singularity at north pole
        heading = 2 * math.atan2(quaternion.x, quaternion.w)
        attitude = math.pi / 2.
        bank = 0.0
    elif test < -0.499 * unit:  # singularity at south pole
        heading = -2 * math.atan2(quaternion.x, quaternion.w)
        attitude = -math.pi / 2.
        bank = 0
    else:
        heading = math.atan2(2 * quaternion.y * quaternion.w - 2 * quaternion.x * quaternion.z, sqx - sqy - sqz + sqw)
        attitude = math.asin(2 * test / unit)
        bank = math.atan2(2 * quaternion.x * quaternion.w - 2 * quaternion.y * quaternion.z, -sqx + sqy - sqz + sqw)
    return Vector3(np.rad2deg([bank, heading, attitude]))


@parameters_as_numpy_arrays('eulers')
def euler_angles_to_quaternion(eulers, dtype=None):
    """Creates a quaternion from a set of Euler angles.

    Rotation order for euler angles is YZX.

    See: http://www.euclideanspace.com/maths/geometry/rotations/conversions/eulerToQuaternion/index.htm

    """
    dtype = dtype or eulers.dtype

    heading_half = eulers[1] / 2.
    attitude_half = eulers[2] / 2.
    bank_half = eulers[0] / 2.

    c1 = np.math.cos(heading_half)
    s1 = np.math.sin(heading_half)
    c2 = np.math.cos(attitude_half)
    s2 = np.math.sin(attitude_half)
    c3 = np.math.cos(bank_half)
    s3 = np.math.sin(bank_half)
    c1c2 = c1 * c2
    s1s2 = s1 * s2
    return Quaternion(
        [
            c1c2 * s3 + s1s2 * c3,
            s1 * c2 * c3 + c1 * s2 * s3,
            c1 * s2 * c3 - s1 * c2 * s3,
            c1c2 * c3 - s1s2 * s3,
        ],
        dtype=dtype
    )



# -*- coding: utf-8 -*-
""" The base of all objects which can be aggregated to describe a 3D scene.
Every Node3D element contains the local and world transformations in 3D space. Every node has one parent node
and a list of child nodes. Every transformation such as, position, rotation, or scale, also affects the
child objects of a node.

The following code example shows how nodes can be created and combined to describe a simple scene with the
following node hierarchy:

Create Hierarchy
    ::

        #     root
        #     /  \\
        #   c_1, c_2
        #       /   \\
        #    c_2_1  c_2_2

        root = Node3D("root")
        c_1 = Node3D("child_1")
        c_2 = Node3D("child_2")
        c_2_1 = Node3D("child_2_1")
        c_2_2 = Node3D("child_2_2")
        root.add(c_1)
        root.add(c_2)
        c_2.add(c_2_1)
        c_2.add(c_2_2)

All Node3D objects can now be individually transformed in 3D space.

Change position of root node
    ::

        root.local_position = Vector([2,1,1])

After executing the code above, all child nodes and the root node itself are moved to the position of the root node.

Change child node local position
    ::

        c_2.local_position += Vector([1,0,0])

The root node and node c_1 are still at position [2,1,1] since they are not affected by the transforms
of the c_2 child node. Both child nodes c_2_1 and c_2_2 and node c_2 itself are shifted 1 unit
towards the local x-axis away from the root node. This means that these nodes are now at
world coordinates [3,1,1].

"""
from copy import copy

from pyrr import Matrix44, Vector3, Quaternion

from pysg.pyrr_extensions import compose_matrix, quaternion_to_euler_angles, euler_angles_to_quaternion
from pysg.util import pyrr_type_checker, parameters_as_angles_deg_to_rad


class Node3D:

    def __init__(self, name: str = "New Node"):
        """  Node element of scene graph (tree structure).

        Args:
            name: Name for string representation.
        """
        self.children = list()
        self._parent = None
        self.name = name

        self._local_position = Vector3()
        self._world_position = Vector3()

        self._scale = Vector3([1., 1., 1.])

        self._local_quaternion = Quaternion()
        self._world_quaternion = Quaternion()
        self._world_matrix = Matrix44.identity()
        self._local_matrix = Matrix44.identity()

        self.__matrix_needs_update = True

    @property
    def parent(self):
        """ The parent of the current node element.

        Returns:
            Node3D: Parent node.

        """
        return self._parent

    @parent.setter
    def parent(self, parent: 'Node3D') -> None:
        self._parent = parent
        if parent is None:
            # If parent is set to None local and world transform are the same
            self._world_position = self._local_position
            self._world_quaternion = self._local_quaternion
        else:
            # If new root node is added the local transform will be set relative to new root node
            self._local_quaternion = self._local_quaternion * parent.world_quaternion.inverse
            self._local_position = self._local_quaternion * (self.world_position - parent.world_position)

    @property
    def local_position(self):
        """ The local position is relative to the transform of the parent node.

        .. note:: The return value is a copy of the original vector and can not be edited directly.
                 This means that code like node.local_position.x += 2 will not work as you might expect it to!

        Returns:
            Vector3: Position of node relative to parent node.

        """
        return copy(self._local_position)

    @local_position.setter
    def local_position(self, local_position: Vector3) -> None:
        local_position = pyrr_type_checker(local_position, Vector3)
        self.__matrix_needs_update = True
        self._local_position = local_position
        world_position_from_rotation = self._world_quaternion * self._local_position
        if self._parent is None:
            self.world_position = world_position_from_rotation
        else:
            self.world_position = self._parent.world_position + world_position_from_rotation

    @property
    def world_position(self):
        """ The world position of a node.

        .. note:: The return value is a copy of the original vector and can not be edited directly.
                 This means that code like node.world_position.x += 2 will not work as you might expect it to!

        Returns:
            Vector3: Position of node in world space.

        """
        return copy(self._world_position)

    @world_position.setter
    def world_position(self, world_position: Vector3) -> None:
        world_position = pyrr_type_checker(world_position, Vector3)
        self.__matrix_needs_update = True
        self._world_position = world_position
        if self._parent is None:
            self._local_position = world_position
        for child in self.children:
            child.world_position = child.local_position + self._world_position

    @property
    def local_quaternion(self):
        """ The local rotation as quaternion.

        .. note:: The return value is a copy of the original quaternion and can not be edited directly.
                Use the quaternion setter instead.

        Returns:
            Quaternion: Quaternion rotation relative to parent node.

        """
        return copy(self._local_quaternion)

    @local_quaternion.setter
    def local_quaternion(self, local_quaternion: Quaternion) -> None:
        local_quaternion = pyrr_type_checker(local_quaternion, Quaternion)
        self.__matrix_needs_update = True
        self._local_quaternion = local_quaternion
        if self._parent is None:
            self.world_quaternion = local_quaternion
        else:
            self.world_quaternion = local_quaternion * self._parent.world_quaternion

    @property
    def world_quaternion(self):
        """ The world rotation as quaternion.

        .. note:: The return value is a copy of the original quaternion and can not be edited directly.
                Use the quaternion setter instead.

        Returns:
            Quaternion: Rotation in world space as quaternion.

        """
        return copy(self._world_quaternion)

    @world_quaternion.setter
    def world_quaternion(self, world_quaternion: Quaternion) -> None:
        world_quaternion = pyrr_type_checker(world_quaternion, Quaternion)
        self.__matrix_needs_update = True
        self._world_quaternion = world_quaternion
        if self._parent is None:
            self._local_quaternion = world_quaternion
        for child in self.children:
            child.world_quaternion = child.world_quaternion * self.world_quaternion

    @property
    def local_euler_angles(self):
        """ Local rotation in euler angle representation as Vector of length 3.
        Internally quaternions are used. The used rotation order is YZX.

        .. seealso:: Please have a look at the :ref:`rotations` section for more details.
        """
        return quaternion_to_euler_angles(self.local_quaternion)

    @local_euler_angles.setter
    @parameters_as_angles_deg_to_rad('euler')
    def local_euler_angles(self, euler: Vector3):
        self.local_quaternion = euler_angles_to_quaternion(euler)

    @property
    def world_euler_angles(self):
        """  World rotation in euler angle representation as Vector of length 3.
        Internally quaternions are used. The used rotation order is YZX.

        .. seealso:: Please have a look at the :ref:`rotations` section for more details.
        """
        return quaternion_to_euler_angles(self.world_quaternion)

    @parameters_as_angles_deg_to_rad('euler')
    @world_euler_angles.setter
    def world_euler_angles(self, euler: Vector3):
        self.world_quaternion = euler_angles_to_quaternion(euler)

    @parameters_as_angles_deg_to_rad('angle')
    def rotate_x(self, angle: float, local_space: bool = True) -> None:
        """ Rotate object around its x-axis.

        Args:
            angle: The rotation angle in degrees.
            local_space: If True rotate in local coordinate system. Otherwise in world space.
        """
        rotation_quaternion = Quaternion.from_x_rotation(angle)
        if local_space:
            self.local_quaternion *= rotation_quaternion
        else:
            self.world_quaternion *= rotation_quaternion

    @parameters_as_angles_deg_to_rad('angle')
    def rotate_y(self, angle: float, local_space: bool = True) -> None:
        """ Rotate object around its y-axis.

        Args:
            angle: The rotation angle in degrees.
            local_space: If True rotate in local coordinate system. Otherwise in world space.
        """
        rotation_quaternion = Quaternion.from_y_rotation(angle)
        if local_space:
            self.local_quaternion *= rotation_quaternion
        else:
            self.world_quaternion *= rotation_quaternion

    @parameters_as_angles_deg_to_rad('angle')
    def rotate_z(self, angle: float, local_space: bool = True) -> None:
        """ Rotate object around its z-axis.

        Args:
            angle: The rotation angle in degrees.
            local_space: If True rotate in local coordinate system. Otherwise in world space.
        """
        rotation_quaternion = Quaternion.from_z_rotation(angle)
        if local_space:
            self.local_quaternion *= rotation_quaternion
        else:
            self.world_quaternion *= rotation_quaternion

    @property
    def local_matrix(self) -> Matrix44:
        """ Local translation matrix (read only).

        Returns:
            Matrix44: 4x4 local translation matrix of the current node.

        """
        if self.__matrix_needs_update:
            self.__matrix_needs_update = False
            self._local_matrix = compose_matrix(self.local_position, self.local_quaternion, self.scale)
        return self._local_matrix

    @property
    def world_matrix(self) -> Matrix44:
        """ World translation matrix (read only).

        Returns:
            Matrix44: 4x4 translation matrix of the current node in world space.

        """
        return self._world_matrix

    @property
    def scale(self) -> Vector3:
        """ Local scale of current node.

        Returns:
            Vector3: X, Y and Z scale as Vector3.

        """
        return copy(self._scale)

    @scale.setter
    def scale(self, scale: Vector3) -> None:
        self.__matrix_needs_update = True
        self._scale = scale

    def __repr__(self):
        return "Node(%s)" % self.name

    def add(self, node_3d: 'Node3D') -> None:
        """ Adds another node as child of this node. It is important to note, that the world position and
        rotation will remain the same of the added child, but all its local transforms will be updated relative
        to the new parent.

        Args:
            node_3d (Node3D): The child node which shall be added to the scene graph.
        """
        self.children.append(node_3d)
        # Local transforms will be updated when setting the parent property
        node_3d.parent = self

    def remove(self, node_3d: 'Node3D') -> None:
        """ Remove a child from the scene graph.

        Args:
            node_3d (Node3D): The child node which shall be removed from the scene graph.
        """
        node_3d.parent = None
        self.children.remove(node_3d)

    def update_world_matrix(self) -> None:
        """ Updates the world matrix for a given node in the render scene graph."""
        if self._parent is None:
            self._world_matrix = self.local_matrix
        else:
            self._world_matrix = self.parent.world_matrix * self.local_matrix

        for child in self.children:
            child.update_world_matrix()

    def get_leaf_nodes(self) -> list:
        """ Recursively iterate over all children and return list with all leaf Node3Ds (no more children).

        Returns:
            list: List of all children with no more child nodes. Type of list elements is Node3D.
        """
        leafs = []

        def _get_leaf_nodes(node):
            if node is not None:
                leafs.append(node)
                for n in node.children:
                    _get_leaf_nodes(n)

        _get_leaf_nodes(self)
        return leafs



import setuptools
import os

with open("README.md", "r", encoding='utf-8') as f:
    long_description = f.read()

dirname = os.path.dirname(__file__)

with open('pysg/version.py', 'r') as f:
    exec(f.read())

setuptools.setup(
    name="pysg",
    version=__version__,
    author="Armin Becher",
    author_email="becherarmin@gmail.com",
    description="Simple and lightweight 3D render scene graph for python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/becheran/pysg",
    packages=setuptools.find_packages(),
    python_requires='>=3.6',
    license='MIT',
    install_requires=[
        'moderngl>=5.4.2,<6',
        'numpy>=1.15.4,<2',
        'pyrr>=0.10.0<1',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    zip_safe=False
)



from unittest import TestCase

from pysg import PerspectiveCamera, OrthographicCamera
from pysg.error import Error


class TestPerspectiveCamera(TestCase):

    def test_parameter_errors_FOV(self):
        with self.assertRaises(Error):
            PerspectiveCamera(fov=-1, aspect=1, near=0.1, far=100)
        with self.assertRaises(Error):
            PerspectiveCamera(fov=0, aspect=1, near=0.1, far=100)
        with self.assertRaises(Error):
            PerspectiveCamera(fov=390, aspect=1, near=0.1, far=100)
        with self.assertRaises(Error):
            PerspectiveCamera(fov=186, aspect=1, near=0.1, far=100)
        with self.assertRaises(Error):
            PerspectiveCamera(fov=180, aspect=1, near=0.1, far=100)

    def test_parameter_errors_aspect_ratio(self):
        with self.assertRaises(Error):
            PerspectiveCamera(fov=45, aspect=-0.1, near=0.1, far=100)
        with self.assertRaises(Error):
            PerspectiveCamera(fov=45, aspect=-1, near=0.1, far=100)
        with self.assertRaises(Error):
            PerspectiveCamera(fov=45, aspect=0.0, near=0.1, far=100)
        with self.assertRaises(Error):
            PerspectiveCamera(fov=45, aspect=0, near=0.1, far=100)

    def test_parameter_errors_near_far(self):
        with self.assertRaises(Error):
            PerspectiveCamera(fov=45, aspect=0.2, near=-1, far=100)
        with self.assertRaises(Error):
            PerspectiveCamera(fov=45, aspect=0.2, near=0, far=0)
        with self.assertRaises(Error):
            PerspectiveCamera(fov=45, aspect=0.2, near=-1, far=1)
        with self.assertRaises(Error):
            PerspectiveCamera(fov=45, aspect=0.2, near=0, far=-1)


class TestOrthographicCamera(TestCase):

    def test_parameter_errors_left_right(self):
        with self.assertRaises(Error):
            OrthographicCamera(left=100, right=-100, top=100, bottom=-100, near=0.1, far=100)
        with self.assertRaises(Error):
            OrthographicCamera(left=0, right=0, top=100, bottom=-100, near=0.1, far=100)
        with self.assertRaises(Error):
            OrthographicCamera(left=100, right=100, top=100, bottom=-100, near=0.1, far=100)

    def test_parameter_errors_top_bottom(self):
        with self.assertRaises(Error):
            OrthographicCamera(left=-100, right=100, top=-100, bottom=100, near=0.1, far=100)
        with self.assertRaises(Error):
            OrthographicCamera(left=-100, right=100, top=100, bottom=100, near=0.1, far=100)
        with self.assertRaises(Error):
            OrthographicCamera(left=-100, right=100, top=0, bottom=0, near=0.1, far=100)

    def test_parameter_errors_near_far(self):
        with self.assertRaises(Error):
            PerspectiveCamera(fov=45, aspect=0.2, near=-1, far=100)
        with self.assertRaises(Error):
            PerspectiveCamera(fov=45, aspect=0.2, near=0, far=0)
        with self.assertRaises(Error):
            PerspectiveCamera(fov=45, aspect=0.2, near=-1, far=1)
        with self.assertRaises(Error):
            PerspectiveCamera(fov=45, aspect=0.2, near=0, far=-1)



""" The camera is used to calculate the viewing frustum in 3D space. All cameras inherit from the Camera class which
itself is of the type Node3D and can be added to the scene graph. """
import math

import pyrr
from pyrr import Matrix44

from pysg.error import ParameterError
from pysg.node_3d import Node3D
from pysg.pyrr_extensions import is_angle


class Camera(Node3D):

    def __init__(self):
        """Base class of all camera types used in pysg."""

        super().__init__()
        self.__projection_matrix = pyrr.Matrix44()
        self._need_matrix_update = True

    @property
    def projection_matrix(self) -> Matrix44:
        """ The current projection matrix of the camera.

        Returns:
            Matrix44: Projection matrix of camera in OpenGL format:\n
            m00 m04 m08 m12\n
            m01 m05 m09 m13\n
            m02 m06 m10 m14\n
            m03 m07 m11 m15
        """
        if self._need_matrix_update:
            self._need_matrix_update = False
            self.__projection_matrix = self._compute_projection_matrix()
        return self.__projection_matrix

    def _compute_projection_matrix(self) -> Matrix44:
        """ Projection matrix calculation based on different camera types.
        Must be implemented by child classes.

        Returns:
            Matrix44: Projection matrix in OpenGL format.

        """
        raise NotImplementedError()


class PerspectiveCamera(Camera):

    def __init__(self, *, fov: float, aspect: float, near: float, far: float):
        """ Camera which uses frustum for the projection matrix.

         Args:
             fov (float): Vertical field of view for the perspective camera in degrees.
             aspect (float): Aspect ratio of camera sensor (with/height).
             near (float): Camera frustum near plane. Everything closer will be culled.
             far (float): Camera frustum far plane. Everything farther away will be culled.

         """
        super().__init__()

        # Check all parameter
        if fov <= 0 \
                or fov >= 180 \
                or not is_angle(fov, in_degrees=True, allow_negative=False, limit_to_circle=True):
            raise ParameterError(fov, 'Field of view with this value is not allowed!')
        if aspect <= 0:
            raise ParameterError(aspect, 'Aspect smaller or equals zero not allowed!')
        if near >= far:
            raise ParameterError((near, far), 'Near must be smaller than far!')
        if near <= 0:
            raise ParameterError(near, 'Near must be greater zero!')
        if far <= 0:
            raise ParameterError(far, 'Far must be greater zero!')

        self.__fov = fov
        self.__aspect = aspect
        self.__near = near
        self.__far = far

    def _compute_projection_matrix(self) -> Matrix44:
        """ Projection matrix calculation for the perspective camera.
            Based on: https://glumpy.github.io/modern-gl.html#projection-matrix
            Only difference is that we use the horizontal FOV not vertical

        Returns:
            Matrix44: The computed projection matrix.
        """

        near_minus_far = self.__near - self.__far
        f = 1 / math.tan(self.__fov / 2)
        aspect_reciprocal = 1 / self.__aspect

        m00 = f
        m11 = f / aspect_reciprocal
        m22 = (self.__far + self.__near) / near_minus_far
        m32 = (2 * self.__near * self.__far) / near_minus_far

        return Matrix44([
            m00, 0.0, 0.0, 0.0,
            0.0, m11, 0.0, 0.0,
            0.0, 0.0, m22, -1,
            0.0, 0.0, m32, 0.0])


class OrthographicCamera(Camera):

    def __init__(self, *, left: float, right: float, top: float, bottom: float, near: float, far: float):
        """  Camera using a box geometry for the projection matrix.

        Args:
            left (float):  Camera volume left (usually -screen_width/2).
            right (float):  Camera volume right (usually screen_width/2).
            top (float):  Camera volume top (usually screen_height/2).
            bottom (float):  Camera volume bottom (usually -screen_height/2)..
            near (float): Camera frustum near plane. Everything closer will be culled.
            far (float): Camera frustum far plane. Everything farther away will be culled.

        """
        super().__init__()

        # Check all parameter
        if left >= right:
            raise ParameterError((left, right), 'Left must be smaller than right!')
        if near >= far:
            raise ParameterError((near, far), 'Near must be smaller than far!')
        if bottom >= top:
            raise ParameterError((top, bottom), 'Bottom must be smaller than top!')
        if near <= 0:
            raise ParameterError(near, 'Near must be greater zero!')
        if far <= 0:
            raise ParameterError(far, 'Far must be greater zero!')

        self.__bottom = bottom
        self.__top = top
        self.__right = right
        self.__left = left
        self.__near = near
        self.__far = far

    def _compute_projection_matrix(self) -> Matrix44:
        """ Projection matrix calculation for the orthographic camera.
                Based on: https://glumpy.github.io/modern-gl.html#projection-matrix
        """

        right_minus_left = self.__right - self.__left
        top_minus_bottom = self.__top - self.__bottom
        far_minus_near = self.__far - self.__near

        m00 = 2 / right_minus_left
        m11 = 2 / top_minus_bottom
        m22 = -2 / far_minus_near

        m30 = -(self.__right + self.__left) / right_minus_left
        m31 = -(self.__top + self.__bottom) / top_minus_bottom
        m32 = -(self.__far + self.__near) / far_minus_near

        return Matrix44([
            m00, 0.0, 0.0, 0.0,
            0.0, m11, 0.0, 0.0,
            0.0, 0.0, m22, 0.0,
            m30, m31, m32, 1.0])
